getmin := proc(L)
	local m,i,k,n;
	n := nops(L);
	m := min(op(L));
	for i to n do
		if L[i] = m then RETURN(i); end if;
	end do;
	RETURN(0);
end:

SimulK := proc(n,params,N,rep,rnge)
	local Y,Kn,B,at,i,j,J,k,r2,epsilon;
	epsilon := 1./(2.*N);
	Y := Matrix(6,rep);
	Kn := Komplete(n);
	r2 := 2*rnge;
	for i to rep do
		B := Flat(Sample(Kn,params,N));
		for j from 1 to r2 do
			J := n-rnge+j;
			at[j] := Allscores(J,Counts(J,0,B,epsilon));
		end do;
		for k to 6 do Y[k,i] := getmin([seq(at[j][k],j=1..r2)])+n-rnge; end do;
	end do;
	RETURN(Y);
end:

SABCK := proc()
#############################################################################
#
#  SABCK(n,params,N,rnge,epsilon(optional))
#
#  
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/(2N) scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,ABC, n,params,N,rnge,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];

	if nargs = 5 then epsilon := args[5];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3];
	end do;
	RETURN(ABC);
end:

SABCK2 := proc()
#############################################################################
#
#  SABCK2(n,params,N,rnge,beta,epsilon(optional))
#
#   beta/(2N) is added to CIC to approximate the missing term...
# 
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/(2N) scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,ABC,b2N, n,params,N,rnge,beta,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];
	beta := args[5];

	if nargs = 6 then epsilon := args[6];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	b2N := beta/(2*N);
	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3]+b2N;
	end do;
	RETURN(ABC);
end:

SABCK3 := proc()
#############################################################################
#
#  SABCK3(n,params,N,rnge,beta(optional),epsilon(optional))
#
#   beta multiplies R/(2N) in CIC to approximate the missing term...
# 
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/N scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,d,ABC,b2N, n,params,N,rnge,beta,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];
	
	if nargs > 4 then beta := args[5];
	else beta := 1;
	end if;

	if nargs = 6 then epsilon := args[6];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	d := 2^n-1;

	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3];
	end do;
	RETURN(ABC);
end:

# n:= 2: N:= 100: rep := 20: rnge := 2: S := SimulK(n,params,N,rep,rnge);
# boxplot(op(convert(S,listlist)),shift = 1,color=red,xtickmarks=0,ytickmarks=[1,2,3,4],labels=["[es, as, res, ras, bic, aic]","n"],title="N=100 samples of K2 repeated rep=20 times\nparams=[.43,.32,.34]");

MInfo2 := proc(t1,t2,t3)
	local p20,p21,r;
	p20 := (1-t2)*(1-t1)+(1-t3)*t1;
	p21 := t2*(1-t1)+t3*t1;
	r := (1-t1)*((1-t2)*log((1-t2)/p20)+t2*log(t2/p21)) +
		t1*((1-t3)*log((1-t3)/p20)+t3*log(t3/p21));
	RETURN(evalf(r));
end:


gfft := proc(X)
#######################################
# X list or vector of n real points
# returns another vector
# with the FT for the real data

	local n,m,j,Z,FTZ;
	n := op(1,convert(X,Vector));
	m := floor(n/2);
	if is(n,even) then
		Z := Vector[row](m,datatype=complex[8]);
		for j to m do
			Z[j] := X[2*j-1] + I*X[2*j];
		end do;
	else 
		Z := Vector[row](m+1,datatype=complex[8]);
		for j to m do
			Z[j] := X[2*j-1] + I*X[2*j];
		end do;
		Z[m+1] := X[n];
	end if;
	FTZ := DiscreteTransforms[FourierTransform](Z);
	RETURN(FTZ);
end:

argmax := proc(X)
	local k,L;
	L := convert(X,list);
	member(max(op(L)),L,'k');
	RETURN(k);
end:

argmin := proc(X)
	local k,L;
	L := convert(X,list);
	member(min(op(L)),L,'k');
	RETURN(k);
end:

plotY := proc()
	local Y,xrange,yrange,
              A,B,C,j,x,y,a,b,m;
	Y := args[1];
	xrange := x=1..10;
	yrange := y=-infinity..infinity;
	if nargs = 3 then 
		xrange := args[2];
		yrange := args[3];
	end if;
	a := op([2,1],xrange);
	b := op([2,2],xrange);
	m := b-a+1;
	A := Y[1,1..m]: B := Y[2,1..m]: C := Y[3,1..m]:
	plot([[seq([j,A[j]],j=1..m)],[seq([j,B[j]],j=1..m)],[seq([j,C[j]],j=1..m)]],
		xrange,yrange,color=[blue,green,red]);
end:	

FindInL := proc(S,L)
##########################################################
# FindInL(S,L)
#   find the positions of the elements of S in the list L
#
#   Returns the list of positions.
##########################################################
	local x,i,j,n,k,J;
	n := nops(L);
	k := nops(S);
	J := [0$k];
	i := 1;
	for x in S do
		for j to n do
			if x = L[j] then
				J[i] := j;
				i := i+1;
				break;
			end if;
		end do;
	end do;
	RETURN(J);
end:

kmins := proc(X,k)
############################################
# return the position of the k consecutive
# mins in X
############################################
	local S,L,J,n;
	L := convert(X,list);
	S := sort(L);
	n := nops(L);
	J := FindInL(S[1..min(n,k)],L);
end:

sng := proc(x)
	if x > 0 then RETURN(1);
	elif x < 0 then RETURN(-1);
	else RETURN(0);
	end if;
end:


VsOf := proc(X)
	local L,i,j,n;
	n := nops(convert(X,list))-1;
	j := 1;
	for i from 2 to n do
		if sng(X[i]-X[i-1]) < sng(X[i+1]-X[i]) then
			L[j] := i;
			j := j+1;
		end if;
	end do;
	RETURN(convert(L,list));
end:

bestC1 := (X,n) -> argmin(X):

SIMUL3 := proc(n,N,rep)
	local i,d,ABC,Y,rnge,m;
	m := 2*n+1;
	rnge := 1..m;
	d := 2^n-1;
	ABC := Matrix(3,rep);
	for i to rep do
		Y := SABCK3(n,[unif(d)],N,rnge);
		ABC[1,i] := argmin(Y[1,rnge]);
		ABC[2,i] := argmin(Y[2,rnge]);
		ABC[3,i] := bestC1(Y[3,rnge],n);
	end do;
	RETURN(ABC);
end:

simulABC := proc(n,N,rep)
	local hists, percents,ABC,Aclass,Bclass,Cclass,Ha,Hb,Hc,j;
	percents := Vector[row](3);
	ABC := SIMUL3(n,N,rep);
	Aclass := stats[transform,tallyinto](convert(ABC[1,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Bclass := stats[transform,tallyinto](convert(ABC[2,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Cclass := stats[transform,tallyinto](convert(ABC[3,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Ha := stats[statplots,histogram](Aclass,color=blue):
	Hb := stats[statplots,histogram](Bclass,color=green):
	Hc := stats[statplots,histogram](Cclass,color=red):
	hists := plots[display]({Ha,Hb,Hc}):
	for j to floor(10/n) do
		percents[1] := 	percents[1] + op(2,Aclass[n*j]);
		percents[2] := 	percents[2] + op(2,Bclass[n*j]);
		percents[3] := 	percents[3] + op(2,Cclass[n*j]);
	end do;
	for j to 3 do percents[j] := evalf(percents[j]/rep); end do;
	RETURN([percents,hists]);
end:

simulABC3 := proc(n,N,rep)
	local hists, percents,ABC,Aclass,Bclass,Cclass,Ha,Hb,Hc,j,m;
#	DEBUG();
	m := 2*n;
	percents := Vector[row](3);
	ABC := SIMUL3(n,N,rep);
	Aclass := stats[transform,tallyinto](convert(ABC[1,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Bclass := stats[transform,tallyinto](convert(ABC[2,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Cclass := stats[transform,tallyinto](convert(ABC[3,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Ha := stats[statplots,histogram](Aclass,color=blue):
	Hb := stats[statplots,histogram](Bclass,color=green):
	Hc := stats[statplots,histogram](Cclass,color=red):
	hists := plots[display]({Ha,Hb,Hc}):
	for j to floor(10/n) do
		percents[1] := 	percents[1] + op(2,Aclass[n*j]);
		percents[2] := 	percents[2] + op(2,Bclass[n*j]);
		percents[3] := 	percents[3] + op(2,Cclass[n*j]);
	end do;
	for j to 3 do percents[j] := evalf(percents[j]/rep); end do;
	RETURN([percents,hists]);
end:
