#Use:> read "c:/cygwin/home/cr569/vol-tests/sampling/vTool1.mpl";

with(LinearAlgebra):

Vols := module()
     export Dag,pa,ch,fillts,Tobin,BlanketOf,b2,b10,posit,vpa,Tprod,ProductProb,Prob,merge,Wi,JP,VolMCMC,VolExact,nparams,DofI,Winaive,FInt,showVolsTool,B,Explode,Collapse,Line,LowVol,HighVol,SCformula,MeanCurvature,LS,fill_g,IntRicci,SortNodes,Sample,Flat,Count,Komplete,unif,Allscores,Escore,Ascore,REscore,RAscore,BIC,AIC,loG,Counts,ABCscores;

local aNext,TableOfPars,ABCscores_old;

Dag := proc(n,ls)
   description "returs an n by n adjacency matrix with links given in the list ls=[i1,j1,i2,j2,...]";
   local i,k,TS;
   global dag,T,S;
   k := nops(ls); if (is(k,odd)) then print("WRONG list of connectors!"); end if;
#   with(LinearAlgebra):
   dag := Matrix(n,n);
   for i from 1 to k by 2 do
       dag[ls[i],ls[i+1]] := 1;
   end do;
   TS := fillts(dag);
   T := TS[1];
   S := TS[2];
   return(dag);
end:

pa := proc(j)
 description "returns the set of parents of node j";
 local s,i;
 s := {};
 for i from 1 to op(1,dag)[1] do
   if dag[i,j]=1 then s := s union {i} end if;
 end do;
return s
end:

ch := proc(i)
  description "returns the set of children of node i";
  local s,j;
  s := {};
  for j from 1 to op(1,dag)[1] do
    if dag[i,j]=1 then s := s union {j} end if;
  end do;
return s
end:

fillts := proc(dag)
 description "returns two vectors: tt=first index of the parameter for each node. st= 2^nops(pa(i))";
 local i,j,tt,st,nnodes,s,np;
 nnodes := op(1,dag)[1]: s := 0: 
 tt := Vector[row](nnodes):
 st := Vector[row](nnodes):
 for i from 1 to nnodes do
    np := nops(pa(i));
    st[i] := 2^np;
    tt[i] := s + 1;
    s := s + st[i];
 end do;
 return tt,st
end:

Tobin := proc(i,k)
  description "transform to binary parents of i =k";
  local j,np,rs,q,r; 
  np := nops(pa(i));
  rs := Vector[row](np); q := k;
  for j from 1 to np do
   q := iquo(q,2,'r');
   rs[np-j+1] := r;
  end do;
  return rs
  end:

BlanketOf := proc(s)
  description "returns the Markov Blanket of the set s";
 local i,j,b;
 b := {};
 for i in s do
  b := b union pa(i)
 end do;
 return b minus s
end:

b2 := proc(k,j)
  description "returns the k least significant binary digits of j";
  local i,q,r,rs;
  rs := Vector[row](k);
  q := j;
  for i from 1 to k do
    q := iquo(q,2,'r');
    rs[k-i+1] := r;
  end do;
  return rs
end:

b10 := proc(v)
  description "returns the base 10 equiv. of the vector of binary digits in v";
  local i,n,x,u; u := convert(v,list);
  x := 0; 
  n := nops(u); 
  for i from 1 to n do
    if (u[i]=1) then x := x+2^(n-i) end if;
  end do;
  return x;
end:

posit := proc(a,s)
  description "returns the position of a in the list s";
  local k;
  member(a,s,'k');
  return k;
end:

vpa := proc(i,s,v)
   description "returns the value of the parents of node i when the set of nodes s has values specified by the binary vector v";
   local par,j,k,r;   
   par := pa(i);
   k := nops(par);
   r := Vector[row](k);
   for j from 1 to k do
     r[j] := v[posit(par[j],s)];
   end do;
   return r
end:

Tprod := proc(i,s,v,T)
  description "returns (1-tn) if ith node has value 0 and tn otherwise. The value of n is computed assuming the set of nodes s has given vector of values v which must include all the parents of i. T is the output of fillts()";
  local n,u,xi; u := vpa(i,s,v); xi := v[posit(i,s)];
  n := T[i] + b10(u);
  if (xi=0) then return (1-t||n) else return t||n end if
end:

ProductProb := proc(s1,s2,v)
  description "returns the product of the probabilities of nodes in the set s1 conditional of their parents when all the values of s1 union s2 are given by the binary vector v. It is assumed that s2 contains the BlanketOf(s1)";
  local i,prod,s; prod := 1;
  s := s1 union s2;
  for i from 1 to nops(s1) do prod := prod*Tprod(s1[i],s,v,T); end do;
  return prod
end:

Prob := proc(s,j)
  description "returns the probability that the nodes in the set s have the values j when written in binary";
  local m,tot,bs,v,vj,vm,k,n; n := nops(s); if (n=0) then return 1 end if;
  bs := BlanketOf(s);
  tot := 0;
  k := nops(bs);
  vj := b2(n,j);
  for m from 0 to 2^k-1 do
    vm := b2(k,m);
    v := merge(s,vj,bs,vm);
    tot := tot + ProductProb(s,bs,v)*Prob(bs,m);
  end do;
  return tot;
end:

merge := proc(s1,v1,s2,v2)
  description "returns the values of s1 union s2 when the nodes in set s1 have values v1 and those in s2 have values v2";
  local i,ns,si,s,v;
  s := s1 union s2;
  ns := nops(s);
  v := Vector[row](ns);
  for i from 1 to ns do
    si := s[i];
    if (si in s1) then v[i] := v1[posit(si,s1)];
      else v[i] := v2[posit(si,s2)];
    end if;
  end do;
  return v;
end:

Wi := proc(i)
  description "returns the contribution of ith node to the det of Fisher info";
  local np,j,tn,s,uprod,dprod;
  s := pa(i);
  np := nops(s); uprod := 1; dprod := 1;
  for j from 0 to 2^np-1 do 
    uprod := uprod*Prob(s,j);
    tn := t || (T[i]+j);
    dprod := dprod*tn*(1-tn);
    end do;
  return uprod/dprod;
end:

JP := proc(dag)
  description "Jeffreys Prior for dag: sqrt(det(I))"; 
  local T,J,i,n,prod;
  T := fillts(dag);
  n := op(1,dag)[1]; prod := 1;
  for i from 1 to n do
    prod := prod*Wi(i);
  end do;
  return sqrt(prod);
end:

VolMCMC := proc(dag,eps)
  description "returns the volume of the dag computed by MCMC with epsilon = eps";
  local i,j,n,s,T,N,prod,nc,np,npis,uprod,tts,childless;
  T := fillts(dag); childless := {};
  n := op(1,dag)[1]; N:= nparams(dag); prod := 1; npis := 0; 
  for i from 1 to n do
    nc := nops(ch(i));
    s := pa(i);   
    np := nops(s);
    if (nc=0) then
        uprod := 1; childless := childless union {i};
        for j from 0 to 2^np-1 do 
        uprod := uprod*Prob(s,j);
        end do;
        prod := prod*uprod;
        npis := npis+2^np;
    else prod := prod*Wi(i);
    end if;
  end do;
  tts := DofI({seq(i,i=1..n)} minus childless);
  return Pi^npis*Int(sqrt(prod),tts,method=_MonteCarlo, epsilon=eps);
end:

VolExact := proc(dag)
  description "returns the volume of the dag as an exact multiple integral";
  local i,j,n,s,T,N,prod,nc,np,npis,uprod,tts,childless;
  T := fillts(dag); childless := {};
  n := op(1,dag)[1]; N:= nparams(dag); prod := 1; npis := 0; 
  for i from 1 to n do
    nc := nops(ch(i));
    s := pa(i);   
    np := nops(s);
    if (nc=0) then
        uprod := 1; childless := childless union {i};
        for j from 0 to 2^np-1 do 
        uprod := uprod*Prob(s,j);
        end do;
        prod := prod*uprod;
        npis := npis+2^np;
    else prod := prod*Wi(i);
    end if;
  end do;
  tts := {seq(i,i=1..n)} minus childless;
  return Pi^npis*FInt(sqrt(prod),tts);
end:


nparams := proc(dag)
  description "returns the total number of parameters of the dag";
  local N,n,i;
  n := op(1,dag)[1]; N := 0;
  for i from 1 to n do 
     N := N + 2^nops(pa(i));     
  end do;
return N;
end:

DofI := proc(s)
   description "returns the domain of integration for all the nodes in the set s";
   local i,j,r;
   r := {};
   for i in s do
      for j from T[i] to T[i]+2^nops(pa(i))-1 do
         r := r union{t||j=0..1};
      end do;
   end do;
return convert(r,'list');
end:

Winaive := proc(i)
  description "returns the contribution of ith node to the det of Fisher info for the naive approximation";
  local np,j,tn,dprod,s,k,sdprod;
  s := pa(i);
  np := nops(s); dprod := 1; k := 2^np;
  for j from 0 to k-1 do 
    tn := t || (T[i]+j);
    dprod := dprod*tn*(1-tn);
    end do; 
  return (1/k)^k/dprod;
end:

FInt := proc(e,s)    
  description "FInt(e,s): returns the multiple integral of the expression e over the set of variables of nodes in s";
     local i,j,tn,r; 
     tn := t||(s[1]);
     r := int(e,tn=0..1);
     for j from T[s[1]] to T[s[1]]+2^nops(pa(s[1]))-1 do
         tn := t||j;
	 r := int(r,tn=0..1);
     end do;
     for i in (s minus {s[1]}) do
         for j from T[i] to T[i]+2^nops(pa(i))-1 do 
               tn := t||j;
	       r := int(r,tn=0..1);
          end do;
      end do;
 return r;
end:

showVolsTool := proc()
   description "Prints all the descriptions in the VolsTool";
   local i;
   print ("Dag(n,ls): returs an n by n adjacency matrix with links given in the list ls=[i1,j1,i2,j2,etc]");

 print ("pa(j): returns the set of parents of node j");

  print ("ch(i): returns the set of children of node i");

 print ("fillts(dag): returns the first index of the parameter for each node");

  print ("Tobin(i,k): transform to binary parents of i =k");

  print ("BlanketOf(s): returns the Markov Blanket of the set s");

  print ("b2(k,j): returns the k least significant binary digits of j");

  print ("b10(v): returns the base 10 equiv. of the vector of binary digits in v");

  print ("posit(a,s): returns the position of a in the list s");

   print ("vpa(i,s,v): returns the value of the parents of node i when the set of nodes s has values specified by the binary vector v");

  print ("Tprod(i,s,v,T): returns (1-tn) if ith node has value 0 and tn otherwise. The value of n is computed assuming the set of nodes s has given vector of values v which must include all the parents of i. T is the output of fillts()");

  print ("ProductProb(s1,s2,v): returns the product of the probabilities of nodes in the set s1 conditional of their parents when all the values of s1 union s2 are given by the binary vector v. It is assumed that s2 contains the BlanketOf(s1)");

  print ("Prob(s,j): returns the probability that the nodes in the set s have the values j when written in binary");

  print ("merge(s1,v1,s2,v2): returns the values of s1 union s2 when the nodes in set s1 have values v1 and those in s2 have values v2");

  print ("Wi(i): returns the contribution of ith node to the det of Fisher info");

  print ("JP(dag): Jeffreys Prior for dag: sqrt(det(I))"); 

  print ("VolMCMC(dag,eps): returns the volume of the dag from MCMC. eps is the desired precission");

  print ("VolExact(dag): returns the volume of the dag as an exact multiple integral");

  print ("LowVol(dag); returns a Lower Bound on the Volume of the dag");

  print ("HighVol(dag); returns an Upper Bound on the Volume of the dag");

  print ("nparams(dag): returns the total number of parameters of the dag");

   print ("DofI(s): returns the domain of integration for all the nodes in the set s");

  print ("Winaive(i): returns the contribution of ith node to the det of Fisher info for the naive approximation");

 print ("pa(j): returns the set of parents of node j");


end:


alias(g=GAMMA):
#B(n) = Beta(n/2+1,n/2+1)=int(t^(n/2)*(1-t)^(n/2),t=0..1);
B := n -> GAMMA(n/2+1)^2/GAMMA(n+2):

# Explode(n) is the volume of the exploding star: one node with n children;
Explode := n -> Pi^(2*n)*B(n-1):

# Collapse(n) is the volume of the collapsing star: one node with n parents;
Collapse := n -> Pi^(2^n)*B(2^(n-1)-1)^n:

# Line(n) is the volume fo the straight line of n nodes;
Line := n -> 4*(Pi/2)^(3*n-4):

LowVol := proc(dag)   
    description  "LowVol(dag); returns a Lower Bound on the Volume of the dag";
    local p,i,j,ai,nnodes;
    nnodes := op(1,dag)[1];
    p := 1;
    for i to nnodes do
        ai := 0;
	for j in ch(i) do
	    ai := ai + (S[j]/S[i]);
        end do; 
        p := p*B(ai/2-1)^S[i];
    end do;
    return evalf(p);
end:

HighVol := proc(dag)    
     description "Upper bound for volume of dag";
     local i,j,bi,nnodes,p,hits,th,C;
     nnodes := op(1,dag)[1];
     hits := Vector[row](nnodes);
     p := 1;
     for i to nnodes do
         bi := 0;
         for j in ch(i) do
             if (pa(i) subset pa(j)) then bi := bi + S[j]/S[i];
             else hits[j] := hits[j]+1;
             end if;
         end do;
         p := p*B(bi/2-1)^S[i];
     end do;
     C := 1; i:=1;
     for i to nnodes do
         if (hits[i]>0) then
	      th := 2^hits[i];
              C := C*(1/th)^(S[i]/2);
         end if;
     end do;
     return evalf(p*C);
end:

fill_g := proc(dag)
  local j,Ti,tk,k;
  global S,T,dim,nnodes,g_compts; 
  for k from 1 to nnodes do
    Ti := T[k];
    for j from 0 to S[k]-1 do
      tk := t||(Ti+j); 
      g_compts[Ti+j,Ti+j] := Prob(pa(k),j)/(tk*(1-tk));
    end do;
  end do;
  eval(g_compts);
end:

SCformula := proc(n,ls)
description "SCformula(n,ls): returns the formula for the Ricci scalar for the dag of n nodes  and with links given in the list ls=[i1,j1,i2,j2,...]";
 local coord,g,ginv,D1g,D2g,Cf1,RMN,RICCI,RS,TS;
 global S,T,dim,nnodes,g_compts,dag;

 nnodes := n;
 dag := Dag(n,ls);
 TS := fillts(dag);
 T := TS[1];
 S := TS[2];
 with(tensor):
 dim := nparams(dag);
 coord := [seq(t||i,i=1..dim)]:
 g_compts := array(symmetric,sparse, 1..dim, 1..dim):
 
 fill_g(dag):
 alias(g = 'g'): 
 g := create( [-1,-1], eval(g_compts));
 ginv := invert( g, 'detg' ):
 D1g := d1metric ( g, coord ):  D2g := d2metric ( D1g, coord ):
 Cf1 := Christoffel1 ( D1g ):
 RMN := Riemann( ginv, D2g, Cf1 ):
 RICCI := Ricci( ginv, RMN ):
 RS := Ricciscalar( ginv, RICCI ):

 RETURN(-RS[compts]);
end:

IntRicci := proc(n,ls)
description "returns the integral of the Ricci scalar over the manifold represented by the dag of n nodes and links in ls=[i1,j1,i2,j2,...]";
  local i,j,s,T,N,prod,nc,np,npis,uprod,tts,childless;
  global dag;

  dag := Dag(n,ls);
  T := fillts(dag); childless := {};
  N:= nparams(dag); prod := 1; npis := 0; 
  for i from 1 to n do
    nc := nops(ch(i));
    s := pa(i);   
    np := nops(s);
    if (nc=0) then
        uprod := 1; childless := childless union {i};
        for j from 0 to 2^np-1 do 
        uprod := uprod*Prob(s,j);
        end do;
        prod := prod*uprod;
        npis := npis+2^np;
    else prod := prod*Wi(i);
    end if;
  end do;
  tts := {seq(i,i=1..n)} minus childless;
  return Pi^npis*FInt(SCformula(n,ls)*sqrt(prod),tts);
end:

MeanCurvature := proc(dag)
local n,ls,i,j;
n := op(1,dag)[1];
ls := LS(dag);
return (IntRicci(n,ls)/VolExact(dag));
end;

LS := proc(dag)
local n,ls,i,j;

ls := [];

n := op(1,dag)[1];
for i to n do
    for j from i+1 to n do
	if (dag[i,j] > 0) then 
             ls := [op(ls),i,j] end if;
    end do;
end do;
return ls;
end;


#######################
# Procedures for sampling
#

unif := stats[random,uniform]:

aNext := proc(S,L)
	local a;
        a := S[1];
	if pa(a) subset convert(L,set) then RETURN(a);
 	else RETURN( aNext( S minus {a}, L ));
        end if;
end:

SortNodes := proc(dag)
############################################################
# SortNodes(dag)
# Input: a dag, i.e. the output of Dag
# Output: L a list of the nodes in the dag sorted
#         so that children are always visited after parents.
############################################################
	local L,S,a,i,j,n;
	n := op(1,dag)[1];
	L := [0$n];
	S := {seq(i, i=1..n)};
	for j to n do
		a := aNext(S,L);
		L[j] := a;
		S := S minus {a};
	end do:
	RETURN(L);
end:

TableOfPars := proc(dag,params)
	local T,i,k,n,xpa,t,tx;
	n := op(1,dag)[1];
	tx := fillts(dag);
        t := tx[1];
	xpa := tx[2];
	T := table();
	for i to n do
		for k from 0 to xpa[i]-1 do	
			T[(i,convert(Tobin(i,k),list))] := params[t[i]+k];
		end do:
	end do:
	RETURN(op(T));
end:

Sample := proc(dag,params,N)
	local i,j,L,theta,n,X,Lpa;
	L := SortNodes(dag);
	theta := TableOfPars(dag,params);
	n := op(1,dag)[1];
	X := Matrix(N,n);
	for j to N do
		for i in L do
			Lpa := convert(X[j,convert(pa(i),list)],list);
			if unif() < theta[(i,Lpa)] then X[j,i] := 1;
			end if;
		end do;
	end do;
	RETURN(X);
end:

Flat := proc(X)
	local B,m,n,i,j,ci;
	m := op(1,X)[1];
	n := op(1,X)[2];
	B := Vector[row](m*n);
	for i to m do
		ci := n*(i-1);
		for j to n do
			if X[i,j] > 0 then B[ci+j] := 1; end if;
		end do;
	end do;
	RETURN(B);
end:

Count := proc(n::posint,offset,B,epsilon)
######################################################################
# Count(n,offset,B)
# Input:
#     n the length of the pattern
# offset < n the starting point in B
#     B a row vector of bits
#   epsilon smoothing parameter > 0 [optional]
#
# Output:
#     A sparse row vector of counts of patterns in B.
######################################################################
	local i,C,j,L,k,p,N;

	N := op(1,B);
	L := convert(B,list);
	k := floor( (N-offset)/n );
	C := Vector[row](2^n,storage=sparse);
	j := offset;
	for i to k do
		p := 1+convert(cat(op(L[j+1..j+n])),decimal,binary);
		C[p] := C[p]+1;
		j := j+n;
	end do;
	RETURN(C);
end:

Counts := proc( )
######################################################################
# Counts(n,offset,B,epsilon)
# Input:
#     n the length of the pattern
# offset < n the starting point in B
#     B a row vector of bits
#   epsilon smoothing parameter > 0 [optional]
#
# Output:
#     A sparse row vector of counts of patterns in B.
######################################################################
	local i,C,j,L,k,p,N,n,offset,B,epsilon;

	n := args[1];
	offset := args[2];
	B := args[3];
	epsilon := 0;
	if nargs = 4 then epsilon := args[4]; end if;
	N := op(1,B);
	L := convert(B,list);
	k := floor( (N-offset)/n );
	C := Vector[row](2^n,storage=sparse);
	j := offset;
	for i to k do
		p := 1+convert(cat(op(L[j+1..j+n])),decimal,binary);
		C[p] := C[p]+1;
		j := j+n;
	end do;
	RETURN(map(`+`,C,epsilon));
end:

Komplete := proc(n)
	local dag,i,j,L;
	L := [];
	for i to n do
		for j from i+1 to n do
			L := [op(L),i,j];
		end do;
	end do;
	RETURN(Dag(n,L));
end:

#### Scoring Procedures

loG := proc(x)
	if evalf(x) > 0 then RETURN(evalf(log(x)));
	else RETURN(0);
	end if;
end:

BIC := proc(n,C)
	local k,d,j,LL,bic,N;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));
	bic := N*loG(N) - LL + (d/2)*loG(N/2/Pi);
	RETURN(evalf(bic));
end:

AIC := proc(n,C)
	local k,d,j,LL,aic,N;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));
	aic := N*loG(N) - LL + d;
	RETURN(evalf(aic));
end:

Escore := proc(n,C)
	local k,d,j,LL,N,escore;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));
	escore := N*loG(N) - LL + (d/2)*loG(N/2/Pi)+ loG(Pi^k/(k-1)!);
	RETURN(evalf(escore));
end:

Ascore := proc(n,C)
	local k,d,j,LL,N,ascore;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));
	ascore := (N+k-1/2)*loG(N) - LL - (1+n*loG(2))*k;
	RETURN(evalf(ascore));
end:

REscore := proc(n,C)
	local d,R,N,j;
	N := convert(C,`+`);
	d := 2^n-1;
	R := d*(d-1)/4;
	RETURN(evalf(Escore(n,C)+R/2/N));
end:

RAscore := proc(n,C)
	local d,R,N,j;
	N := convert(C,`+`);
	d := 2^n-1;
	R := d*(d-1)/4;
	RETURN(evalf(Ascore(n,C)+R/2/N));
end:

Allscores := proc(n,C)
	local k,d,j,LL,N,R2N,NLN,d2L,es,as,res,ras,bic,aic;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));

	NLN := N*loG(N);
	d2L := (d/2)*loG(N/2/Pi);

	as := (N+k-1/2)*loG(N) - LL - (1+n*loG(2))*k;
	es := NLN - LL + d2L + loG(Pi^k/(k-1)!);

	R2N := d*(d-1)/8/N;
	ras := as + R2N;
	res := es + R2N;

        aic := NLN - LL + d;
	bic := NLN - LL + d2L;
	RETURN([es,as,res,ras,bic,aic]);
end:

ABCscores := proc(n,C)
################################################################
# Uses the new Rterm
#
#   Rterm = Pi*R/N/log(d+1) = 4.532*R/N/n = 1.133*d*(d-1)/n/N
#
################################################################
	local k,d,dN,j,LL,N,Rterm,d2L,LogV,bic,aic,cic;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);

	LL := evalf(N*loG(N) - sum('C[j]*loG(C[j])','j'=1..(d+1)));
	dN := evalf(d);
	d2L := evalf((d/2)*loG(N/2/Pi));
	Rterm := evalf(1.133*d*(d-1)/n/N); # notice this is 4.532*R/n/N
	LogV := evalf(loG(Pi^k/(k-1)!));

        aic := LL + dN;
	bic := LL + d2L;
	cic := LL + d2L + LogV + Rterm;
	RETURN([aic,bic,cic]);
end:

ABCscores_old := proc(n,C)
	local k,d,j,LL,N,R2N,NLN,d2L,bic,aic,cic;
	d := 2^n-1;
	k := 2^(n-1);
	N := convert(C,`+`);
	LL := sum('C[j]*loG(C[j])','j'=1..(d+1));

	NLN := N*loG(N);
	d2L := (d/2)*loG(N/2/Pi);
	R2N := d*(d-1)/8/N;

        aic := NLN - LL + d;
	bic := NLN - LL + d2L;
	cic := NLN - LL + d2L + loG(Pi^k/(k-1)!) + R2N;
	RETURN([aic,bic,cic]);
end:

end module;

with(Vols):

getmin := proc(L)
	local m,i,k,n;
	n := nops(L);
	m := min(op(L));
	for i to n do
		if L[i] = m then RETURN(i); end if;
	end do;
	RETURN(0);
end:

SimulK := proc(n,params,N,rep,rnge)
	local Y,Kn,B,at,i,j,J,k,r2,epsilon;
	epsilon := 1./(2.*N);
	Y := Matrix(6,rep);
	Kn := Komplete(n);
	r2 := 2*rnge;
	for i to rep do
		B := Flat(Sample(Kn,params,N));
		for j from 1 to r2 do
			J := n-rnge+j;
			at[j] := Allscores(J,Counts(J,0,B,epsilon));
		end do;
		for k to 6 do Y[k,i] := getmin([seq(at[j][k],j=1..r2)])+n-rnge; end do;
	end do;
	RETURN(Y);
end:

SABCK := proc()
#############################################################################
#
#  SABCK(n,params,N,rnge,epsilon(optional))
#
#  
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/(2N) scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,ABC, n,params,N,rnge,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];

	if nargs = 5 then epsilon := args[5];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3];
	end do;
	RETURN(ABC);
end:

SABCK2 := proc()
#############################################################################
#
#  SABCK2(n,params,N,rnge,beta,epsilon(optional))
#
#   beta/(2N) is added to CIC to approximate the missing term...
# 
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/(2N) scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,ABC,b2N, n,params,N,rnge,beta,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];
	beta := args[5];

	if nargs = 6 then epsilon := args[6];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	b2N := beta/(2*N);
	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3]+b2N;
	end do;
	RETURN(ABC);
end:

SABCK3 := proc()
#############################################################################
#
#  SABCK3(n,params,N,rnge,beta(optional),epsilon(optional))
#
#   beta multiplies R/(2N) in CIC to approximate the missing term...
# 
description "	SABCK(n,params,N,rep,rnge,epsilon(optional)) \n Takes N samples from Komplete(n) and computes the AIC,BIC and CIC = BIC+log(Vol)+R/N scores on these data assuming n= rnge where rnge is a range like 1..8 \n The optional parameter epsilon controls the degree of smoothing of observed counts (adds epsilon to each count), it defaults to 1/(2N)";

	local Kn,B,at,i,j,J,k,Y,r,a,b,d,ABC,b2N, n,params,N,rnge,beta,epsilon;

	n := args[1];
	params := args[2];
	N := args[3];
	rnge := args[4];
	
	if nargs > 4 then beta := args[5];
	else beta := 1;
	end if;

	if nargs = 6 then epsilon := args[6];
	  else epsilon := 1./(2.*N);
	end if;

	Kn := Komplete(n);
	B := Flat(Sample(Kn,params,N));

	a := op(1,rnge);
	b := op(2,rnge);
	r := b-a+1;
	ABC := Matrix(3,r);
	d := 2^n-1;

	J := 0;
	for i from a to b do
		J := J + 1;
		Y := ABCscores(i,Counts(i,0,B,epsilon));
		ABC[1,J] := Y[1];
		ABC[2,J] := Y[2];
		ABC[3,J] := Y[3];
	end do;
	RETURN(ABC);
end:

# n:= 2: N:= 100: rep := 20: rnge := 2: S := SimulK(n,params,N,rep,rnge);
# boxplot(op(convert(S,listlist)),shift = 1,color=red,xtickmarks=0,ytickmarks=[1,2,3,4],labels=["[es, as, res, ras, bic, aic]","n"],title="N=100 samples of K2 repeated rep=20 times\nparams=[.43,.32,.34]");

MInfo2 := proc(t1,t2,t3)
	local p20,p21,r;
	p20 := (1-t2)*(1-t1)+(1-t3)*t1;
	p21 := t2*(1-t1)+t3*t1;
	r := (1-t1)*((1-t2)*log((1-t2)/p20)+t2*log(t2/p21)) +
		t1*((1-t3)*log((1-t3)/p20)+t3*log(t3/p21));
	RETURN(evalf(r));
end:


gfft := proc(X)
#######################################
# X list or vector of n real points
# returns another vector
# with the FT for the real data

	local n,m,j,Z,FTZ;
	n := op(1,convert(X,Vector));
	m := floor(n/2);
	if is(n,even) then
		Z := Vector[row](m,datatype=complex[8]);
		for j to m do
			Z[j] := X[2*j-1] + I*X[2*j];
		end do;
	else 
		Z := Vector[row](m+1,datatype=complex[8]);
		for j to m do
			Z[j] := X[2*j-1] + I*X[2*j];
		end do;
		Z[m+1] := X[n];
	end if;
	FTZ := DiscreteTransforms[FourierTransform](Z);
	RETURN(FTZ);
end:

argmax := proc(X)
	local k,L;
	L := convert(X,list);
	member(max(op(L)),L,'k');
	RETURN(k);
end:

argmin := proc(X)
	local k,L;
	L := convert(X,list);
	member(min(op(L)),L,'k');
	RETURN(k);
end:

plotY := proc()
	local Y,xrange,yrange,
              A,B,C,j,x,y,a,b,m;
	Y := args[1];
	xrange := x=1..10;
	yrange := y=-infinity..infinity;
	if nargs = 3 then 
		xrange := args[2];
		yrange := args[3];
	end if;
	a := op([2,1],xrange);
	b := op([2,2],xrange);
	m := b-a+1;
	A := Y[1,1..m]: B := Y[2,1..m]: C := Y[3,1..m]:
	plot([[seq([j,A[j]],j=1..m)],[seq([j,B[j]],j=1..m)],[seq([j,C[j]],j=1..m)]],
		xrange,yrange,color=[blue,green,red]);
end:	

FindInL := proc(S,L)
##########################################################
# FindInL(S,L)
#   find the positions of the elements of S in the list L
#
#   Returns the list of positions.
##########################################################
	local x,i,j,n,k,J;
	n := nops(L);
	k := nops(S);
	J := [0$k];
	i := 1;
	for x in S do
		for j to n do
			if x = L[j] then
				J[i] := j;
				i := i+1;
				break;
			end if;
		end do;
	end do;
	RETURN(J);
end:

kmins := proc(X,k)
############################################
# return the position of the k consecutive
# mins in X
############################################
	local S,L,J,n;
	L := convert(X,list);
	S := sort(L);
	n := nops(L);
	J := FindInL(S[1..min(n,k)],L);
end:

sng := proc(x)
	if x > 0 then RETURN(1);
	elif x < 0 then RETURN(-1);
	else RETURN(0);
	end if;
end:


VsOf := proc(X)
	local L,i,j,n;
	n := nops(convert(X,list))-1;
	j := 1;
	for i from 2 to n do
		if sng(X[i]-X[i-1]) < sng(X[i+1]-X[i]) then
			L[j] := i;
			j := j+1;
		end if;
	end do;
	RETURN(convert(L,list));
end:

bestC1 := (X,n) -> argmin(X):

SIMUL3 := proc(n,N,rep)
	local i,d,ABC,Y,rnge,m;
	m := 2*n+1;
	rnge := 1..m;
	d := 2^n-1;
	ABC := Matrix(3,rep);
	for i to rep do
		Y := SABCK3(n,[unif(d)],N,rnge);
		ABC[1,i] := argmin(Y[1,rnge]);
		ABC[2,i] := argmin(Y[2,rnge]);
		ABC[3,i] := bestC1(Y[3,rnge],n);
	end do;
	RETURN(ABC);
end:

simulABC := proc(n,N,rep)
	local hists, percents,ABC,Aclass,Bclass,Cclass,Ha,Hb,Hc,j;
	percents := Vector[row](3);
	ABC := SIMUL3(n,N,rep);
	Aclass := stats[transform,tallyinto](convert(ABC[1,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Bclass := stats[transform,tallyinto](convert(ABC[2,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Cclass := stats[transform,tallyinto](convert(ABC[3,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..18)]):
	Ha := stats[statplots,histogram](Aclass,color=blue):
	Hb := stats[statplots,histogram](Bclass,color=green):
	Hc := stats[statplots,histogram](Cclass,color=red):
	hists := plots[display]({Ha,Hb,Hc}):
	for j to floor(10/n) do
		percents[1] := 	percents[1] + op(2,Aclass[n*j]);
		percents[2] := 	percents[2] + op(2,Bclass[n*j]);
		percents[3] := 	percents[3] + op(2,Cclass[n*j]);
	end do;
	for j to 3 do percents[j] := evalf(percents[j]/rep); end do;
	RETURN([percents,hists]);
end:

simulABC3 := proc(n,N,rep)
	local hists, percents,ABC,Aclass,Bclass,Cclass,Ha,Hb,Hc,j,m;
#	DEBUG();
	m := 2*n;
	percents := Vector[row](3);
	ABC := SIMUL3(n,N,rep);
	Aclass := stats[transform,tallyinto](convert(ABC[1,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Bclass := stats[transform,tallyinto](convert(ABC[2,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Cclass := stats[transform,tallyinto](convert(ABC[3,1..rep],list),
                          [seq(i+0.5..i+1.5,i=0..m)]):
	Ha := stats[statplots,histogram](Aclass,color=blue):
	Hb := stats[statplots,histogram](Bclass,color=green):
	Hc := stats[statplots,histogram](Cclass,color=red):
	hists := plots[display]({Ha,Hb,Hc}):
	for j to floor(10/n) do
		percents[1] := 	percents[1] + op(2,Aclass[n*j]);
		percents[2] := 	percents[2] + op(2,Bclass[n*j]);
		percents[3] := 	percents[3] + op(2,Cclass[n*j]);
	end do;
	for j to 3 do percents[j] := evalf(percents[j]/rep); end do;
	RETURN([percents,hists]);
end:

rev := ListTools:-Reverse:
Bhex := proc(a,b)
### the following function Bhex(a,b) returns the bytes numbered a to b in 
### hexadecimal with the bits obtained from the Vector of bits B
	local n0,n1;
	if a = 0 then n0 := 1;
	else n0 := a*8;
	end if;
	n1 := b*8;
        map(convert,rev(convert(rev(convert(
             B[n0..n1],list)),base,2,16)),hex);
end:

nmax := 10:
movewindow := proc(wsize,winc,B)
	local N,n,j,cic,nf,epsilon,pj; #DEBUG();
	global nmax;
	epsilon := 0;
	N := op(1,B); # number of bits in B
	nf := floor((N-wsize)/winc); # number of frames
	cic := Matrix(nf,nmax);
	for n to nmax do
		cic[1,n] := ABCscores(n,CountX(n,B[1..wsize]),epsilon)[3];
	end do;
	pj := wsize;  # previous j
	for j from 2 to nf do 
		for n to nmax do
			cic[j,n] := ABCscores(n,CountX(n,B[pj+1..pj+wsize]),epsilon)[3];
		end;
		pj := pj+winc;
	end;
	RETURN(cic);
end:

mwin1 := proc(wsize,winc,B)
	local N,n,j,cic,nf,epsilon,pj,C; #DEBUG();
	global nmax;
	epsilon := 0;
	N := op(1,B); # number of bits in B
	nf := floor((N-wsize)/winc); # number of frames
	cic := Matrix(nf,nmax);
	C := table();
	for n to nmax do
		C[1,n] := CountX(n,B[1..wsize]);
		cic[1,n] := ABCscores(n,C[1,n],epsilon)[3];
	end do;
	pj := wsize;  # previous j
	for j from 2 to nf do 
		for n to nmax do
			C[j,n] := CountX(n,B[pj+1..pj+wsize]);
			cic[j,n] := ABCscores(n,C[j,n],epsilon)[3];
		end;
		pj := pj+winc;
	end;
	RETURN([cic,C]);
end:

P := t -> plot([seq([j,CIC[floor(t),j]],j=1..nmax)],x=1..nmax,y=0..500):
plotmovieCIC := CIC -> plots[animate](P,[t],t=1..op([1,1],CIC),frames=op([1,1],CIC)):

P1 := (t,CIC) -> plot([seq([j,CIC[floor(t),j]],j=1..nmax)],x=1..nmax,y=0..500):
pmovieCIC := CIC -> plots[animate](P1,[t,CIC],t=1..op([1,1],CIC),frames=op([1,1],CIC)):




Bytes2bits := proc(L)
####################################################################
#
#  Bytes2bits(L) converts a list of bytes in decimal to a row Vector
#                of bits
#
#  e.g. fd := fopen("image.jpg",READ,BINARY);
#       B := Bytes2bits(readbytes(fd,5000));
####################################################################
	local i,n,B,b,k,i8,j;
	n := nops(L);
	B := Vector[row](8*n,datatype=integer[1]);
	for i to n do
		b := convert(convert(L[i],binary),string);
		k := length(b);
		if k < 8 then b := convert(cat(0$(8-k),b),string); end if;
		i8 := 8*(i-1);
		for j to 8 do
			if b[j] = "1" then B[i8+j] := 1; end if;
		end do;
	end do;
	RETURN(B);
end:



CountX := proc()
######################################################################
# CountX(n,offset (optional),B)  get the counts for xtra large n (n>30).
#
# Input:
#     n the length of the pattern
#  (optional) offset < n the starting point in B
#     B a row vector of bits
#
#
# Output:
#     A sparse row vector of counts of patterns in B.
######################################################################
	local i,C,j,L,k,p,N,T,n,offset,B;
	offset := 0;
	n := args[1];
	if nargs = 2 then 
		B := args[2];
	elif nargs = 3 then 
		offset := args[2];
		B := args[3];
	end if;
		
	N := op(1,B);
	L := convert(B,list);
	k := floor( (N-offset)/n );
	T := table([]);
	j := offset;
	for i to k do
		p := convert(cat(op(L[j+1..j+n])),decimal,binary);
#		p := convert(p,hex);
		if is(T[p],integer) then
			T[p] := T[p]+1;
		else T[p] := 1;
		end if;
		j := j+n;
	end do;
	C := [2^n,convert(op(op(T)),set)];
	RETURN(C);
end:


LLandN := proc()
#######################################################################
#
#  LLandN(C,epsilon(optional))
#
#  In: C = counts
#      epsilon = (optional) smoothing parameter. Defaults to 0.
#                 If epsilon < 0 then is defined as 1/d/N
#
#  Out: [-LogLikelihood, N]
########################################################################
	local C,epsilon,j,d,n0,N,Ne,s,ll,e;

	epsilon := 0;
	C := args[1];
	N := 0;
	for e in op(2,C) do
		N := N+op(2,e);
	end do;

	s := 0.; d := op(1,C)-1;

	if nargs = 2 then 
		epsilon := args[2]; 
		if epsilon < 0 then epsilon := evalf(1/d/N); end if;
	end if;
	
	
	if epsilon > 0 then
		for e in op(2,C) do	
			j := op(2,e);
			s := s + evalf((j+epsilon)*log(j+epsilon));
		end do;
		n0 := op(1,C) - nops(op(2,C));
		Ne := evalf(N+(d+1)*epsilon);
		ll := evalf(Ne*log(Ne)-s - n0*epsilon*log(epsilon));
	else
		for e in op(2,C) do	
			j := op(2,e);
			if j <> 1 then 	s := s + evalf(j*log(j)); end if;
		end do;
		ll := evalf(N*log(N) - s);
	end if;
	RETURN([ll,N]);
end:

ABCscores := proc()
#####################################################################
#
#  ABCscores(n(optional),C,epsilon(optional))
#
#  In:  n = (optional) number of nodes in Komplete(n)
#       C = sparse vector of counts the result of Counts(n,offset,B)
#       epsilon = (optional) smoothing parameter for counts,
#                  defaults to 0. If epsilon < 0 then is taken as
#                  1/d/N.
#  Out: AIC, BIC and CIC based on the counts in C.
####################################################################
	local k,d,dN,j,LL,LLN,N,R2N,d2L,LogV,bic,aic,cic,n,C,epsilon;

	epsilon := 0;
	if nargs = 1 then 
		C := args[1]; 
		d := op(1,C)-1;
		k := (d+1)/2;
		n := simplify(log[2](d+1));
	elif nargs > 1 then 
		n := args[1];
		d := 2^n-1;
		k := (d+1)/2;
		C := args[2];
	end if;

	if nargs = 3 then epsilon := args[3]; end if;

	LLN := LLandN(C,epsilon);
	LL := LLN[1];
	N := LLN[2];

	dN := evalf(d);
	d2L := evalf((d/2)*log(N/2/Pi));
	R2N := evalf((10/n)*d*(d-1)/8/N); # notice this is  (10/n)*R/(2N)

	if n >= 30 then 
		LogV := evalf( k*log(Pi) - k*log(k) - k - 0.5*log(2*Pi) );
	else
		LogV := evalf( k*log(Pi) - log(GAMMA(k)));
	end if;

        aic := LL + dN;
	bic := LL + d2L;
	cic := LL + d2L + LogV + R2N;
	RETURN([aic,bic,cic,LL]);
end:

b2hex := proc(L::list(posint))
## transf. list of bytes (in decimal) to hex.
## Each hex byte is a TWO hex character like A7 or 04 or 0D.
## Makes sure the initial 0 in bytes like 04 don't disapear!
   local j,H;
   H := "";
   for j to nops(L) do
       if L[j] < 16 then H := cat(H,"0",convert(L[j],hex,decimal));
       else H := cat(H,convert(L[j],hex,decimal));
       end if;
   end do;
   RETURN(H);
end:
s2n := proc(str::string)
# string to number conversion
   local kb,k16,k10;
   kb := convert(str,bytes);
   k16 := b2hex(kb);
   k10 := convert(k16,decimal,hex);
end:

n2s := proc(n::posint)
# number to string conversion
   local hexn, L,j,decn; 
   L := "";
   hexn := convert(convert(n,hex,decimal),string);
   if type(length(hexn),odd) then hexn := cat("0",hexn); end if;
   for j from 1 by 2 to length(n) do
       decn := convert(hexn[j..(j+1)],decimal,hex);
       L := cat(L,convert([decn],bytes));
   end do;
   RETURN(L);
end:

s2b := proc(s)
##############################################
# s2b(s) : string to bits
	local B,n,N,s1,V,j,r;
	s1 := convert(convert(s2n(s),binary),string);
	n := length(s1);
	r := irem(n,8);
	if r <> 0 then 
		s1 := cat("0"$(8-r),s1); 
		n := n+8-r;
	end if;
	V := Vector[row](n,datatype=integer[1]);
	for j to n do
		if s1[j] = "1" then V[j] := 1; end if;
	end do;
	RETURN(V);
end:


scoreBits := proc (B,rangen) 
	local r, ABC, n, J, epsilon, a, b, Y; 
#	DEBUG(); 
	epsilon := 0; 
	a := op(1, rangen); 
	b := op(2, rangen); 
	r := b-a+1; 
	ABC := Matrix(3, r); 
	J := 0; 
	for n from a to b do 
		J := J+1; 
		Y := ABCscores(CountX(n, 0, B));
		ABC[1, J] := Y[1]; 
		ABC[2, J] := Y[2]; 
		ABC[3, J] := Y[3] 
	end do; 
	RETURN(ABC) 
end:

KL := proc()
########################################################################
#
# KL(C1,C2,(optional)epsilon)
#    Returns the Kullback number between P(C1) and P(C2)
#     where C1 and C2 are observed counts,i.e., the output of CountX
#     epsilon >=0 is an optional smoothing parameter for the zero counts.
#     It defaults to 10^(-100) (almost no smoothing)
######################################################################### 
	local C1,C2,epsilon,j,d,s,mj,nj,M,N,TotM,TotN,r,e;
	C1 := args[1];
	C2 := args[2];
	epsilon := 0;
	if nargs = 3 then epsilon := args[3]; end if;
	if epsilon <= 0 then epsilon := evalf(10^(-100));end if;
	d := C1[1]-1; 
	s := 0.;
	M := table([op(C1[2])]);
	N := table([op(C2[2])]);
	TotM := (d+1)*epsilon;
	TotN := (d+1)*epsilon;
	for e in C1[2] do
		j := lhs(e);
		mj := M[j];
		nj := N[j];
		if (not is(nj,integer)) and  is(mj,integer) then
			s := s + (mj+epsilon)*log((mj+epsilon)/epsilon);
			TotM := TotM + mj;
		elif is(nj,integer) and (not is(mj,integer)) then
			s := s + epsilon*log(epsilon/(nj+epsilon));
			TotN := TotN + nj;
		elif is(nj,integer) and is(mj,integer) then
			s := s + (mj+epsilon)*log((mj+epsilon)/(nj+epsilon));
			TotM := TotM + mj;
			TotN := TotN + nj;
		end if;
	end do;
	for e in C2[2] do
		j := lhs(e);
		if is(M[j],integer) then 
			;
		else # i.e. mj := 0;
			nj := N[j];
			s := s + epsilon*log(epsilon/(nj+epsilon));
			TotN := TotN + nj;
		end if;
	end do;

	r := evalf( s/TotM - log(TotM/TotN) );
	RETURN([r,TotM,TotN]);
end:

b2B := proc(B)
########################################################################
#
#  b2B(B) Transforms a Vector B of bits into a List of Bytes
#
########################################################################
	local n,i,j,r,Bytes,b,k,s,p,rev;
	rev := ListTools:-Reverse;
	n := iquo(op(1,B),8,'r');
	if r > 0 then
		Bytes := Vector[row](n+1);
	else Bytes := Vector[row](n);
	end if;
	p := 0;
	for j from 1 to n do
		b := convert(rev(convert(B[p+1..p+8],list)),base,2,10);
		k := nops(b);
		s := 0;
		for i to k do s := s + b[i]*10^(i-1); end do;
		Bytes[j] := s;
		p := p+8;
	end do;
	# the reminder byte if any
	if r > 0 then
		b := convert(rev(convert(B[p+1..p+r],list)),base,2,10);
		k := nops(b);
		s := 0;
		for i to k do s := s + b[i]*10^(i-1); end do;
		Bytes[n+1] := s;
		p := p+8;
	end if;
	RETURN(Bytes);
end:
